from service_capacity_modeling.capacity_planner import planner
from service_capacity_modeling.interface import CapacityDesires
from service_capacity_modeling.interface import certain_int
from service_capacity_modeling.interface import DataShape
from service_capacity_modeling.interface import Interval
from service_capacity_modeling.interface import QueryPattern


user_desires = CapacityDesires(
    service_tier=0,
    query_pattern=QueryPattern(
        estimated_read_per_second=certain_int(100000),
        estimated_write_per_second=certain_int(100000),
        estimated_mean_read_size_bytes=Interval(
            low=10, mid=100, high=1000, confidence=0.98
        ),
    ),
    data_shape=DataShape(
        estimated_state_size_gib=certain_int(10),
    ),
)


def test_cassandra_merge():
    cass_defaults = planner.models["org.netflix.cassandra"].default_desires(
        user_desires, {}
    )
    merged = user_desires.merge_with(cass_defaults)

    # Should come from the user
    assert merged.service_tier == 0
    assert merged.query_pattern.estimated_read_per_second.mid == 100000
    assert merged.query_pattern.estimated_mean_read_size_bytes.low == 10
    assert merged.data_shape.estimated_state_size_gib.mid == 10

    # Should come from cassandra
    assert merged.query_pattern.estimated_mean_read_latency_ms.mid == 2.0
    assert merged.query_pattern.estimated_mean_write_latency_ms.mid == 1.0

    # Should come from overall defaults
    assert merged.core_reference_ghz == 2.3

    # Should come from the default size producing the count
    # 10 GiB / 512 byte items = 20971520 items
    assert merged.data_shape.estimated_state_item_count is not None
    assert int(merged.data_shape.estimated_state_item_count.mid) == 41943040



from service_capacity_modeling.interface import Instance
from service_capacity_modeling.tools.auto_shape import reshape_family


def test_reshape_family():
    result = reshape_family("r6id.2xlarge")

    expected_large = Instance.model_validate(
        {
            "name": "r6id.large",
            "cpu": 2,
            "cpu_ghz": 3.5,
            "ram_gib": 15.45,
            "net_mbps": 781.25,
            "drive": {
                "name": "ephem",
                "size_gib": 110,
                "read_io_latency_ms": {
                    "minimum_value": 0.05,
                    "low": 0.10,
                    "mid": 0.125,
                    "high": 0.17,
                    "maximum_value": 2.00,
                    "confidence": 0.9,
                },
                "read_io_per_s": 33542,
                "write_io_per_s": 16771,
                "block_size_kib": 4,
                "single_tenant": False,
            },
        }
    )
    assert result[0].name == "r6id.large"
    assert result[0] == expected_large

    expected_24xlarge = Instance.model_validate(
        {
            "name": "r6id.24xlarge",
            "cpu": 96,
            "cpu_ghz": 3.5,
            "ram_gib": 741.6,
            "net_mbps": 37500.0,
            "drive": {
                "name": "ephem",
                "size_gib": 5292,
                "read_io_latency_ms": {
                    "minimum_value": 0.05,
                    "low": 0.10,
                    "mid": 0.125,
                    "high": 0.17,
                    "maximum_value": 2.05,
                    "confidence": 0.9,
                },
                "read_io_per_s": 1610004,
                "write_io_per_s": 805008,
                "block_size_kib": 4,
                "single_tenant": True,
            },
        }
    )

    assert result[-1].name == "r6id.24xlarge"
    assert result[-1] == expected_24xlarge


def test_reshape_with_override():
    result = reshape_family("r6id.2xlarge")

    for shape in result:
        assert shape.drive is not None
        if shape.name == "r6id.large":
            # This shape has a hard coded override on the drive
            assert shape.drive.read_io_latency_ms.maximum_value == 2.0
        else:
            assert shape.drive.read_io_latency_ms.maximum_value == 2.05



from .instances import get_simple_instance_headroom_target
from .instances import get_simple_instance_headroom_target_for_name

__all__ = [
    "get_simple_instance_headroom_target",
    "get_simple_instance_headroom_target_for_name",
]



from service_capacity_modeling.hardware import shapes
from service_capacity_modeling.simple import get_simple_instance_headroom_target
from service_capacity_modeling.simple import (
    get_simple_instance_headroom_target_for_name,
)


def test_invalid_instance_returns_none():
    assert get_simple_instance_headroom_target_for_name("invalid.instance") is None


def test_m5_same_headroom_as_r5():
    m5_headroom = get_simple_instance_headroom_target_for_name("m5.xlarge")
    assert m5_headroom is not None
    r5_headroom = get_simple_instance_headroom_target_for_name("r5.xlarge")
    assert r5_headroom is not None
    assert m5_headroom == r5_headroom


def test_larger_instance_needs_less_headroom():
    xlarge_headroom = get_simple_instance_headroom_target_for_name("m5.xlarge")
    assert xlarge_headroom is not None
    twoxlarge_headroom = get_simple_instance_headroom_target_for_name("m5.2xlarge")
    assert twoxlarge_headroom is not None
    assert twoxlarge_headroom < xlarge_headroom


def test_has_some_headroom():
    m5_instance = shapes.hardware.regions["us-east-1"].instances["m5.xlarge"]
    r5_instance = shapes.hardware.regions["us-east-1"].instances["r5.xlarge"]
    assert m5_instance.cpu == r5_instance.cpu
    assert m5_instance.cpu == 4
    m5_headroom = get_simple_instance_headroom_target(m5_instance)
    r5_headroom = get_simple_instance_headroom_target(m5_instance)
    assert m5_headroom is not None
    assert r5_headroom is not None
    assert m5_headroom > 0
    assert m5_headroom == r5_headroom



import logging
import math
import random
from decimal import Decimal
from typing import Callable
from typing import List
from typing import Optional
from typing import Tuple

from service_capacity_modeling.interface import AVG_ITEM_SIZE_BYTES
from service_capacity_modeling.interface import CapacityDesires
from service_capacity_modeling.interface import CapacityPlan
from service_capacity_modeling.interface import certain_float
from service_capacity_modeling.interface import certain_int
from service_capacity_modeling.interface import Clusters
from service_capacity_modeling.interface import Drive
from service_capacity_modeling.interface import Instance
from service_capacity_modeling.interface import Interval
from service_capacity_modeling.interface import RegionClusterCapacity
from service_capacity_modeling.interface import RegionContext
from service_capacity_modeling.interface import Requirements
from service_capacity_modeling.interface import ServiceCapacity
from service_capacity_modeling.interface import ZoneClusterCapacity
from service_capacity_modeling.models import utils

logger = logging.getLogger(__name__)


SECONDS_IN_YEAR = 31556926


# In square root staffing we have to take into account the QOS parameter
# Which is related to the probability that a user queues. On low tier clusters
# (aka critical clusters) we want a lower probability of queueing
def _QOS(tier: int) -> float:
    # Halfin-Whitt delay function
    # P(queue) ~= [1 + B * normal_cdf(B) / normal_pdf(B)] ^ -1
    #
    # P(queue) ~= 0.01
    if tier == 0:
        return 2.375
    # P(queue) ~= 0.05
    elif tier == 1:
        return 1.761
    # P(queue) ~= 0.2
    elif tier == 2:
        return 1.16
    # P(queue) ~= 0.29 ~= 0.3
    else:
        return 1


def _sqrt_staffed_cores(rps: float, latency_s: float, qos: float) -> int:
    # Square root staffing
    # s = a + Q*sqrt(a)
    return int(math.ceil((rps * latency_s) + qos * math.sqrt(rps * latency_s)))


def sqrt_staffed_cores(desires: CapacityDesires) -> int:
    """Computes cores given a sqrt staffing model

    Little's Law: Concurrency = Average Rate * Average Latency
    For example: 0.1 average concurrency = 100 / second * 1 millisecond

    However, if you provision for average, when statistically unlikely traffic
    spikes happen, you will queue, creating _latency_.

    Square root staffing says to avoid that latency instead of provisioning
    average number of cores, you provision

    Cores = (Rate * Latency) + (QoS * sqrt(Rate * Latency))
    Cores = (Required cores) + (Safety margin)

    Pick higher QoS to minimize the probability of queueing. In our case we do it
    based on tier.
    """
    qos = _QOS(desires.service_tier)
    read_rps, read_lat = (
        desires.query_pattern.estimated_read_per_second.mid,
        desires.query_pattern.estimated_mean_read_latency_ms.mid / 1000.0,
    )
    write_rps, write_lat = (
        desires.query_pattern.estimated_write_per_second.mid,
        desires.query_pattern.estimated_mean_write_latency_ms.mid / 1000.0,
    )

    read_cores = _sqrt_staffed_cores(read_rps, read_lat, qos)
    write_cores = _sqrt_staffed_cores(write_rps, write_lat, qos)

    return read_cores + write_cores


def simple_network_mbps(desires: CapacityDesires) -> int:
    """Computes network mbps with a simple model"""
    read_bytes_per_second = (
        desires.query_pattern.estimated_read_per_second.mid
        * desires.query_pattern.estimated_mean_read_size_bytes.mid
    )
    write_bytes_per_second = (
        desires.query_pattern.estimated_write_per_second.mid
        * desires.query_pattern.estimated_mean_write_size_bytes.mid
    )

    net_bytes_per_sec = read_bytes_per_second + write_bytes_per_second

    return int(max(1, math.ceil(net_bytes_per_sec / 125000)))


def network_services(
    service_type: str,
    context: RegionContext,
    desires: CapacityDesires,
    copies_per_region: int,
) -> List[ServiceCapacity]:
    result = []
    # Network transfer is for every other zone and then for every region
    # other than us as well.
    num_zones = max(copies_per_region - 1, 0)
    num_regions = max(context.num_regions - 1, 0)

    # have bytes and / second
    size = desires.query_pattern.estimated_mean_write_size_bytes.mid
    wps = desires.query_pattern.estimated_write_per_second.mid
    # need gib and / year

    txfer_gib = (wps * size / (1024 * 1024 * 1024)) * (SECONDS_IN_YEAR)

    # For each cross region replication we have to pay to move bytes
    # inter region. This is the number of regions minus 1
    inter_txfer = context.services.get("net.inter.region", None)
    if inter_txfer:
        result.append(
            ServiceCapacity(
                service_type=f"{service_type}.net.inter.region",
                annual_cost=(inter_txfer.annual_cost_gib(txfer_gib) * num_regions),
                service_params={"txfer_gib": txfer_gib, "num_regions": num_regions},
            )
        )

    # Same zone is free, but we pay for replication from our zone to others
    intra_txfer = context.services.get("net.intra.region", None)
    if intra_txfer:
        result.append(
            ServiceCapacity(
                service_type=f"{service_type}.net.intra.region",
                annual_cost=(
                    intra_txfer.annual_cost_gib(txfer_gib)
                    * num_zones
                    * context.num_regions
                ),
                service_params={
                    "txfer_gib": txfer_gib,
                    "num_zones": num_zones,
                    "num_regions": context.num_regions,
                },
            )
        )
    return result


def compute_stateless_region(  # pylint: disable=too-many-positional-arguments
    instance: Instance,
    needed_cores: int,
    needed_memory_gib: float,
    needed_network_mbps: float,
    # Faster CPUs can execute operations faster
    core_reference_ghz: float,
    num_zones: int = 3,
) -> RegionClusterCapacity:
    """Computes a regional cluster of a stateless app

    Basically just takes into cpu, memory, and network

    returns: (count of instances, annual cost in dollars)
    """

    # Stateless apps basically just use CPU resources and network
    needed_cores = math.ceil(
        max(1, needed_cores // (instance.cpu_ghz / core_reference_ghz))
    )

    count = max(2, math.ceil(needed_cores / instance.cpu))

    # Now take into account the network bandwidth
    count = max(count, math.ceil(needed_network_mbps / instance.net_mbps))

    # Now take into account the needed memory
    count = max(count, math.ceil(needed_memory_gib / instance.ram_gib))

    # Try to keep zones balanced
    count = utils.next_n(count, num_zones)

    return RegionClusterCapacity(
        cluster_type="stateless-app",
        count=count,
        instance=instance,
        annual_cost=count * instance.annual_cost,
    )


# pylint: disable=too-many-locals
# pylint: disable=too-many-statements
def compute_stateful_zone(  # pylint: disable=too-many-positional-arguments
    instance: Instance,
    drive: Drive,
    needed_cores: int,
    needed_disk_gib: float,
    needed_memory_gib: float,
    needed_network_mbps: float,
    # Faster CPUs can execute operations faster
    core_reference_ghz: float,
    # Cloud drives may need to scale for IOs, and datastores might need more
    # or less IOs for a given data size as well as space
    # Contract for disk ios is
    # (per_node_size_gib, node_count) -> (read_ios, write_ios)
    required_disk_ios: Callable[
        [float, int], Tuple[float, float]
    ] = lambda size_gib, count: (0, 0),
    required_disk_space: Callable[[float], float] = lambda size_gib: size_gib,
    # The maximum amount of state we can hold per node in the database
    # typically you don't want stateful systems going much higher than a
    # few TiB so that recovery functions properly
    max_local_disk_gib: float = 2048,
    # Some stateful clusters have sidecars that take memory
    reserve_memory: Callable[[float], float] = lambda x: 0,
    # How much write buffer we get per instance (usually a percentage of
    # the reserved memory, e.g. for buffering writes in heap)
    write_buffer: Callable[[float], float] = lambda x: 0,
    required_write_buffer_gib: float = 0,
    # Some stateful clusters have preferences on per zone sizing
    cluster_size: Callable[[int], int] = lambda x: x,
    min_count: int = 0,
    adjusted_disk_io_needed: float = 0.0,
    read_write_ratio: float = 0.0,
) -> ZoneClusterCapacity:
    # Normalize the cores of this instance type to the latency reference
    needed_cores = math.ceil(
        max(1, needed_cores // (instance.cpu_ghz / core_reference_ghz))
    )

    # Datastores often require disk headroom for e.g. compaction and such
    if instance.drive is not None:
        needed_disk_gib = math.ceil(required_disk_space(needed_disk_gib))

    # How many instances do we need for the CPU
    count = math.ceil(needed_cores / instance.cpu)

    # How many instances do we need for the ram, taking into account
    # reserved memory for the application and system
    count = max(
        count,
        math.ceil(
            needed_memory_gib / (instance.ram_gib - reserve_memory(instance.ram_gib))
        ),
    )
    # Account for if the stateful service needs a certain amount of reserved
    # memory for a given throughput.
    if write_buffer(instance.ram_gib) > 0:
        count = max(
            count,
            math.ceil(required_write_buffer_gib / (write_buffer(instance.ram_gib))),
        )

    # How many instances do we need for the network
    count = max(count, math.ceil(needed_network_mbps / instance.net_mbps))

    # How many instances do we need for the disk
    if (
        instance.drive is not None
        and instance.drive.size_gib > 0
        and max_local_disk_gib > 0
    ):
        disk_per_node = min(max_local_disk_gib, instance.drive.size_gib)
        count = max(count, math.ceil(needed_disk_gib / disk_per_node))
        if adjusted_disk_io_needed != 0.0:
            instance_read_iops = (
                instance.drive.read_io_per_s
                if instance.drive.read_io_per_s is not None
                else 0
            )
            assert isinstance(instance_read_iops, int)
            instance_write_iops = (
                instance.drive.write_io_per_s
                if instance.drive.write_io_per_s is not None
                else 0
            )
            assert isinstance(instance_write_iops, int)
            instance_adjusted_io = (
                (
                    read_write_ratio * float(instance_read_iops)
                    + (1.0 - read_write_ratio) * float(instance_write_iops)
                )
                * instance.drive.block_size_kib
                * 1024.0
            )
            if instance_adjusted_io != 0.0:
                count = max(
                    count, math.ceil(adjusted_disk_io_needed / instance_adjusted_io)
                )

    count = max(cluster_size(count), min_count)
    cost = count * instance.annual_cost

    attached_drives = []
    if instance.drive is None and required_disk_space(needed_disk_gib) > 0:
        # If we don't have disks attach the cloud drive with enough
        # space and IO for the requirement

        # Note that cloud drivers are provisioned _per node_ and must be chosen for
        # the max of space and IOS.
        space_gib = max(1, math.ceil(required_disk_space(needed_disk_gib) / count))
        read_io, write_io = required_disk_ios(space_gib, count)
        read_io, write_io = (
            utils.next_n(read_io, n=200),
            utils.next_n(write_io, n=200),
        )
        total_ios = read_io + write_io
        io_gib = cloud_gib_for_io(drive, total_ios, space_gib)

        # Provision EBS in increments of 100 GiB
        ebs_gib = utils.next_n(max(1, io_gib, space_gib), n=100)

        # When initially provisioniong we don't want to attach more than
        # 1/3 the maximum volume size in one node (preferring more nodes
        # with smaller volumes)
        max_size = drive.max_size_gib / 3
        if ebs_gib > max_size > 0:
            ratio = ebs_gib / max_size
            count = max(cluster_size(math.ceil(count * ratio)), min_count)
            cost = count * instance.annual_cost
            ebs_gib = max_size

        read_io, write_io = required_disk_ios(space_gib, count)
        read_io, write_io = (
            utils.next_n(read_io, n=200),
            utils.next_n(write_io, n=200),
        )
        if (read_io + write_io) > drive.max_io_per_s:
            ratio = (read_io + write_io) / drive.max_io_per_s
            count = max(cluster_size(math.ceil(count * ratio)), min_count)
            cost = count * instance.annual_cost
            read_io = utils.next_n(read_io * ratio, n=200)
            write_io = utils.next_n(write_io * ratio, n=200)

        attached_drive = drive.model_copy()
        attached_drive.size_gib = ebs_gib
        attached_drive.read_io_per_s = int(round(read_io, 2))
        attached_drive.write_io_per_s = int(round(write_io, 2))

        # TODO: appropriately handle RAID setups for throughput requirements
        attached_drives.append(attached_drive)

        cost = cost + (attached_drive.annual_cost * count)

    logger.debug(
        "For (cpu, memory_gib, disk_gib) = (%s, %s, %s) need (%s, %s, %s, %s)",
        needed_cores,
        needed_memory_gib,
        needed_disk_gib,
        count,
        instance.name,
        attached_drives,
        cost,
    )

    return ZoneClusterCapacity(
        cluster_type="stateful-cluster",
        count=count,
        instance=instance,
        attached_drives=attached_drives,
        annual_cost=cost,
    )


# AWS GP2 gives 3 IOS / gb stored.
def gp2_gib_for_io(read_ios) -> int:
    return int(max(1, read_ios // 3))


def cloud_gib_for_io(drive, total_ios, space_gib) -> int:
    if drive.name == "gp2":
        return gp2_gib_for_io(total_ios)
    else:
        return space_gib


class WorkingSetEstimator:
    def __init__(self):
        self._cache = {}

    def working_set_percent(
        self,
        # latency distributions of the read SLOs versus the drives
        # expressed as scipy rv_continuous objects
        drive_read_latency_dist,
        read_slo_latency_dist,
        # what percentile of disk latency should we target for keeping in
        # memory. Not as this is _increased_ more memory will be reserved
        target_percentile: float = 0.90,
        min_working_set: float = 0.01,
    ) -> Interval:
        # random cache eviction
        if len(self._cache) >= 100:
            self._cache.pop(random.choice(tuple(self._cache.keys())))

        cache_key = (
            id(drive_read_latency_dist),
            id(read_slo_latency_dist),
            target_percentile,
        )
        # Cached because ppf in particular is _really_ slow
        if cache_key not in self._cache:
            # How fast is the drive at the target percentile
            minimum_drive_latency = drive_read_latency_dist.ppf(target_percentile)

            # How much of the read latency SLO lies below the minimum
            # drive latency. So for example if EBS's 99% is 1.7ms and we
            # 45% of our read SLO lies below that then we need at least 45%
            # of our data to be stored in memory.
            required_percent = float(read_slo_latency_dist.cdf(minimum_drive_latency))

            self._cache[cache_key] = certain_float(
                max(required_percent, min_working_set)
            )
        return self._cache[cache_key]


_working_set_estimator = WorkingSetEstimator()


def working_set_from_drive_and_slo(
    # latency distributions of the read SLOs versus the drives
    # expressed as scipy rv_continuous objects
    drive_read_latency_dist,
    read_slo_latency_dist,
    estimated_working_set: Optional[Interval] = None,
    # what percentile of disk latency should we target for keeping in
    # memory. Not as this is _increased_ more memory will be reserved
    target_percentile: float = 0.90,
    min_working_set: float = 0.01,
) -> Interval:
    if estimated_working_set is not None:
        return estimated_working_set

    return _working_set_estimator.working_set_percent(
        drive_read_latency_dist=drive_read_latency_dist,
        read_slo_latency_dist=read_slo_latency_dist,
        target_percentile=target_percentile,
        min_working_set=min_working_set,
    )


def item_count_from_state(
    estimated_state_size_gib: Interval,
    estimated_state_item_count: Optional[Interval] = None,
) -> Interval:
    if estimated_state_item_count is not None:
        return estimated_state_item_count

    return certain_int(
        int(estimated_state_size_gib.mid * 1024 * 1024 * 1024) // AVG_ITEM_SIZE_BYTES
    )


def _add_optional_float(
    left: Optional[float], right: Optional[float]
) -> Optional[float]:
    if left is None and right is None:
        return None
    if left is None:
        return right
    if right is None:
        return left
    return left + right


def _add_interval(left: Interval, right: Interval) -> Interval:
    return Interval(
        low=(left.low + right.low),
        mid=(left.mid + right.mid),
        high=(left.high + right.high),
        confidence=min(left.confidence, right.confidence),
        model_with=left.model_with,
        minimum_value=_add_optional_float(left.minimum_value, right.minimum_value),
        maximum_value=_add_optional_float(left.maximum_value, right.maximum_value),
    )


def _noop_zone(x: ZoneClusterCapacity) -> ZoneClusterCapacity:
    return x


def _noop_region(x: RegionClusterCapacity) -> RegionClusterCapacity:
    return x


def merge_requirements(
    left_req: Requirements,
    right_req: Requirements,
) -> Requirements:
    merged_zonal, merged_regional = [], []
    for req in list(left_req.zonal) + list(right_req.zonal):
        merged_zonal.append(req)
    for req in list(left_req.regional) + list(right_req.regional):
        merged_regional.append(req)

    merged_regrets = set(left_req.regrets) | set(right_req.regrets)

    return Requirements(
        zonal=merged_zonal, regional=merged_regional, regrets=tuple(merged_regrets)
    )


def merge_plan(
    left: Optional[CapacityPlan],
    right: Optional[CapacityPlan],
    zonal_transform: Callable[[ZoneClusterCapacity], ZoneClusterCapacity] = _noop_zone,
    regional_transform: Callable[
        [RegionClusterCapacity], RegionClusterCapacity
    ] = _noop_region,
) -> Optional[CapacityPlan]:
    if left is None or right is None:
        return None

    merged_requirements = merge_requirements(left.requirements, right.requirements)

    left_cluster = left.candidate_clusters
    right_cluster = right.candidate_clusters

    merged_annual_costs = {}
    all_sources = set(
        left_cluster.annual_costs.keys() | right_cluster.annual_costs.keys()
    )

    merged_annual_costs = {
        k: (
            left_cluster.annual_costs.get(k, Decimal(0))
            + right_cluster.annual_costs.get(k, Decimal(0))
        )
        for k in all_sources
    }

    merged_clusters = Clusters(
        annual_costs=merged_annual_costs,
        zonal=(
            [zonal_transform(z) for z in left_cluster.zonal]
            + [zonal_transform(z) for z in right_cluster.zonal]
        ),
        regional=(
            [regional_transform(z) for z in left_cluster.regional]
            + [regional_transform(z) for z in right_cluster.regional]
        ),
        services=(list(left_cluster.services) + list(right_cluster.services)),
    )
    return CapacityPlan(
        requirements=merged_requirements, candidate_clusters=merged_clusters
    )



import argparse
import json
import sys
from collections import OrderedDict
from fractions import Fraction
from typing import Optional
from typing import Sequence

from service_capacity_modeling.hardware import shapes
from service_capacity_modeling.interface import Drive
from service_capacity_modeling.interface import Instance


def pull_family(
    instance_filter: str = "m7a", region: str = "us-east-1"
) -> Sequence[Instance]:
    # pylint: disable=unused-argument
    """TODO: Talk to the AWS API

    This will emulate

    aws --region us-east-1 \
      ec2 describe-instance-types \
      --filters 'Name=instance-type,Values=m7a*'

    And then convert from base 10 for memory and disk to base 2 and generate the
    shape from that.

    Returns a list of Instance for hardware. Then we can compose to writing
    json files.
    """
    return []


def _scale_drive(drive: Optional[Drive], scale: Fraction) -> Optional[Drive]:
    if drive is None:
        return None
    # This is wrong but ... not too wrong
    single_tenant = False
    if scale > 4:
        single_tenant = True

    read_iops, write_iops = drive.read_io_per_s, drive.write_io_per_s
    if read_iops is not None:
        read_iops = int(round(read_iops * scale))
    if write_iops is not None:
        write_iops = int(round(write_iops * scale))

    return Drive(
        name=drive.name,
        size_gib=int(drive.size_gib * scale),
        read_io_latency_ms=drive.read_io_latency_ms,
        read_io_per_s=read_iops,
        write_io_per_s=write_iops,
        block_size_kib=drive.block_size_kib,
        single_tenant=single_tenant,
    )


def reshape_family(
    initial_shape: str, reference_family: str = "m5", region: str = "us-east-1"
) -> Sequence[Instance]:
    """Takes a single instance as a template and generates the other shapes

    Useful for filling in details missing from automated pull such as IOPS
    """
    shape = shapes.hardware.regions[region].instances[initial_shape]
    family, normalized_size = shape.family, shape.normalized_size

    normalized_sizes = [
        i.normalized_size
        for i in shapes.hardware.regions[region].instances.values()
        if i.family == reference_family
    ]

    def name(input_size: Fraction) -> str:
        if input_size.denominator == 1 and input_size.numerator != 1:
            size = str(input_size.numerator) + "xlarge"
        else:
            size = {
                Fraction(1, 4): "medium",
                Fraction(1, 2): "large",
                Fraction(1, 1): "xlarge",
            }[input_size]
        return f"{family}{shape.family_separator}{size}"

    new_shapes = []
    for new_size in normalized_sizes:
        if normalized_size == new_size:
            continue
        scale_factor = new_size / normalized_size

        current_shape = shapes.hardware.regions[region].instances.get(name(new_size))

        new_shape = Instance(
            name=name(new_size),
            cpu=int(shape.cpu * scale_factor),
            cpu_ghz=shape.cpu_ghz,
            ram_gib=round(shape.ram_gib * scale_factor, 2),
            net_mbps=round(shape.net_mbps * scale_factor, 2),
            drive=_scale_drive(shape.drive, scale_factor),
        )
        # When we have layers, reverse this and clean it up
        merged = new_shape
        if current_shape is not None:
            merged = new_shape.merge_with(current_shape)
        merged_dict = merged.model_dump(exclude_unset=True)
        if "annual_cost" in merged_dict:
            del merged_dict["annual_cost"]
        if "drive" in merged_dict and "annual_cost" in merged_dict["drive"]:
            del merged_dict["drive"]["annual_cost"]

        new_shapes.append(Instance(**merged_dict))

    return sorted(new_shapes, key=lambda i: i.normalized_size)


def main(args) -> int:
    for template in args.template:
        result = reshape_family(template, args.reference_family, args.region)
        print("Hardware shapes for aws.json")
        json_shapes = OrderedDict()
        for shape in result:
            model_dict = shape.model_dump(exclude_unset=True)
            # Hardware shouldn't have costs
            if "annual_cost" in model_dict:
                del model_dict["annual_cost"]
            if "drive" in model_dict and "annual_cost" in model_dict["drive"]:
                del model_dict["drive"]["annual_cost"]

            json_shapes[shape.name] = model_dict

        print(f"[{template}[aws.json] Hardware Shapes", file=sys.stderr)
        print(json.dumps(json_shapes, indent=2))

    return 0


if __name__ == "__main__":
    instances = set()
    families = set()
    regions = set()
    for ar in shapes.hardware.regions.keys():
        regions.add(ar)
        for instance in shapes.hardware.regions[ar].instances.values():
            instances.add(instance.name)
            families.add(instance.family)

    parser = argparse.ArgumentParser(
        prog="Project shapes from template",
        description="Takes one verified shape and projects for rest",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument("-f", "--reference-family", choices=families, default="m5")
    parser.add_argument("-r", "--region", choices=regions, default="us-east-1")
    parser.add_argument("template", nargs="+", choices=instances)
    sys.exit(main(parser.parse_args()))



from typing import Optional

from service_capacity_modeling.hardware import shapes
from service_capacity_modeling.interface import Instance

FULL_CORE_INSTANCE_FAMILIES = ["c7a", "m7a", "r7a"]


def _family_from_instance_type(instance_type: str) -> str:
    return instance_type.split(".")[0]


def _headroom_approx(cores: int, is_ht: bool) -> float:
    """For implementation see /notebooks/headroom-estimator.ipynb"""

    # Hyperthreading performance penalty
    # When hyperthreading is enabled, each core is not as effective as a physical core
    # We estimate each virtual core to be about 60% as effective as a physical core
    HT_PENALTY = 0.6

    # Adjust effective cores if hyperthreading is enabled
    # This accounts for the reduced effectiveness of virtual cores
    cores_f = float(cores)
    if is_ht:
        cores_f = cores_f * HT_PENALTY

    # Calculate required headroom using magic formula:
    return 0.712 / (cores_f**0.448)


def _is_instance_type_hyperthreads(instance_type: str) -> bool:
    instance_family = _family_from_instance_type(instance_type)
    return instance_family not in FULL_CORE_INSTANCE_FAMILIES


def get_simple_instance_headroom_target(instance: Instance) -> Optional[float]:
    """Determine an approximate headroom target for an instance given its
    instance type.

    The headroom target should be the percentage of CPU that should be
    reserved for headroom to ensure sensible performance profile.

    This could be 1-utilization_target, however we leave the ultimate
    utilization_target to the caller, since, we do not know how much
    operational headroom they want to leave (ie: success buffer).

    For example, a response here of "headroom = 15%", means caller could
    decide with a success_buffer=1 to use a utilization_target of 85%.
    For success_buffer>1, they should target below 85% utilization.

    This is only suitable for "single-thread-like" workloads, which
    fortunately many stateless services are.

    For implementation see /notebooks/headroom-estimator.ipynb
    """
    try:
        is_ht = _is_instance_type_hyperthreads(instance.name)
        core_count = instance.cpu
        return _headroom_approx(core_count, is_ht)
    except (KeyError, AttributeError):
        return None


def get_simple_instance_headroom_target_for_name(instance_type: str) -> Optional[float]:
    try:
        instance = shapes.hardware.regions["us-east-1"].instances[instance_type]
        return get_simple_instance_headroom_target(instance)
    except (KeyError, AttributeError):
        return None
